﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using HeuristicLab.Algorithms.DataAnalysis.SymRegGrammarEnumeration.GrammarEnumeration;
using HeuristicLab.Common;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Problems.DataAnalysis.Symbolic;

namespace HeuristicLab.Algorithms.DataAnalysis.SymRegGrammarEnumeration {
  public class Grammar {

    public Symbol StartSymbol;

    #region Symbols

    public VariableSymbol Var;

    public NonterminalSymbol Expr;
    public NonterminalSymbol Term;
    public NonterminalSymbol Factor;
    public NonterminalSymbol LogFactor;
    public NonterminalSymbol ExpFactor;
    public NonterminalSymbol SinFactor;

    public NonterminalSymbol SimpleExpr;
    public NonterminalSymbol SimpleTerm;

    public TerminalSymbol Addition;
    public TerminalSymbol Multiplication;
    public TerminalSymbol Log;
    public TerminalSymbol Exp;
    public TerminalSymbol Sin;

    // For infix notation
    public TerminalSymbol OpeningBracket;
    public TerminalSymbol ClosingBracket;

    #endregion


    #region HL Symbols for Parsing ExpressionTrees

    private TypeCoherentExpressionGrammar symbolicExpressionGrammar;

    private ISymbol constSy;
    private ISymbol varSy;

    private ISymbol addSy;
    private ISymbol mulSy;
    private ISymbol logSy;
    private ISymbol expSy;
    private ISymbol divSy;
    private ISymbol sinSy;

    private ISymbol rootSy;
    private ISymbol startSy;

    #endregion

    public Grammar(string[] variables) {
      #region Define Symbols
      Var = new VariableSymbol("var", variables);

      Expr = new NonterminalSymbol("Expr");
      Term = new NonterminalSymbol("Term");
      Factor = new NonterminalSymbol("Factor");
      LogFactor = new NonterminalSymbol("LogFactor");
      ExpFactor = new NonterminalSymbol("ExpFactor");
      SinFactor = new NonterminalSymbol("SinFactor");

      SimpleExpr = new NonterminalSymbol("SimpleExpr");
      SimpleTerm = new NonterminalSymbol("SimpleTerm");

      Addition = new TerminalSymbol("+");
      Multiplication = new TerminalSymbol("*");
      Log = new TerminalSymbol("log");
      Exp = new TerminalSymbol("exp");
      Sin = new TerminalSymbol("sin");

      OpeningBracket = new TerminalSymbol("(");
      ClosingBracket = new TerminalSymbol(")");
      #endregion


      #region Production rules
      // order of production is important, since they are accessed via index
      // in memoization.
      StartSymbol = Expr;

      Expr.AddProduction(Term, Expr, Addition);
      Expr.AddProduction(Term);

      Term.AddProduction(Factor, Term, Multiplication);
      Term.AddProduction(Factor);

      Factor.AddProduction(Var);
      Factor.AddProduction(LogFactor);
      Factor.AddProduction(ExpFactor);
      Factor.AddProduction(SinFactor);

      LogFactor.AddProduction(SimpleExpr, Log);
      ExpFactor.AddProduction(SimpleTerm, Exp);
      SinFactor.AddProduction(SimpleExpr, Sin);

      SimpleExpr.AddProduction(SimpleTerm, SimpleExpr, Addition);
      SimpleExpr.AddProduction(SimpleTerm);

      SimpleTerm.AddProduction(Var, SimpleTerm, Multiplication);
      SimpleTerm.AddProduction(Var);
      #endregion

      #region Parsing to SymbolicExpressionTree
      symbolicExpressionGrammar = new TypeCoherentExpressionGrammar();
      symbolicExpressionGrammar.ConfigureAsDefaultRegressionGrammar();

      constSy = symbolicExpressionGrammar.Symbols.OfType<Constant>().First();
      varSy = symbolicExpressionGrammar.Symbols.OfType<Variable>().First();
      addSy = symbolicExpressionGrammar.Symbols.OfType<Addition>().First();
      mulSy = symbolicExpressionGrammar.Symbols.OfType<Multiplication>().First();
      logSy = symbolicExpressionGrammar.Symbols.OfType<Logarithm>().First();
      expSy = symbolicExpressionGrammar.Symbols.OfType<Exponential>().First();
      divSy = symbolicExpressionGrammar.Symbols.OfType<Division>().First();
      sinSy = symbolicExpressionGrammar.Symbols.OfType<Sine>().First();

      rootSy = symbolicExpressionGrammar.Symbols.OfType<ProgramRootSymbol>().First();
      startSy = symbolicExpressionGrammar.Symbols.OfType<StartSymbol>().First();

      #endregion
    }

    #region Hashing
    public int CalcHashCode(SymbolString sentence) {
      Debug.Assert(sentence.Any(), "Trying to evaluate empty sentence!");
      // Debug.Assert(sentence.All(s => s is TerminalSymbol), "Trying to evaluate symbol sequence with nonterminalsymbols!");

      Stack<Symbol> parseStack = new Stack<Symbol>(sentence);

      Symbol peek = parseStack.Peek();
      int[] subtreeHashes = GetSubtreeHashes(parseStack);
      return AggregateHashes(peek, subtreeHashes);
    }

    private int[] GetSubtreeHashes(Stack<Symbol> parseStack) {
      Symbol currentSymbol = parseStack.Pop();

      // MULTIPLICATION
      if (ReferenceEquals(currentSymbol, Multiplication)) {
        List<int> childHashes = new List<int>();

        // First subtree
        if (ReferenceEquals(parseStack.Peek(), Multiplication)) {
          childHashes.AddRange(GetSubtreeHashes(parseStack));
        } else {
          childHashes.Add(AggregateHashes(parseStack.Peek(), GetSubtreeHashes(parseStack)));
        }
        // Second subtree
        if (ReferenceEquals(parseStack.Peek(), Multiplication)) {
          childHashes.AddRange(GetSubtreeHashes(parseStack));
        } else {
          childHashes.Add(AggregateHashes(parseStack.Peek(), GetSubtreeHashes(parseStack)));
        }

        // Sort due to commutativity
        childHashes.Sort();
        return childHashes.ToArray();
      }

      // ADDITION
      if (ReferenceEquals(currentSymbol, Addition)) {
        HashSet<int> uniqueChildHashes = new HashSet<int>();

        // First subtree
        if (ReferenceEquals(parseStack.Peek(), Addition)) {
          uniqueChildHashes.UnionWith(GetSubtreeHashes(parseStack));
        } else {
          var peek = parseStack.Peek();
          uniqueChildHashes.Add(AggregateHashes(peek, GetSubtreeHashes(parseStack)));
        }
        // Second subtree
        if (ReferenceEquals(parseStack.Peek(), Addition)) {
          uniqueChildHashes.UnionWith(GetSubtreeHashes(parseStack));
        } else {
          var peek = parseStack.Peek();
          uniqueChildHashes.Add(AggregateHashes(peek, GetSubtreeHashes(parseStack)));
        }

        var result = uniqueChildHashes.ToList();
        result.Sort();
        return result.ToArray();
      }

      // LOG, EXP, SIN
      if (ReferenceEquals(currentSymbol, Log) || ReferenceEquals(currentSymbol, Exp) ||
          ReferenceEquals(currentSymbol, Sin)) {
        return AggregateHashes(parseStack.Peek(), GetSubtreeHashes(parseStack)).ToEnumerable().ToArray();
      }

      // var or nonterminal symbol
      return currentSymbol.StringRepresentation.GetHashCode().ToEnumerable().ToArray();
    }

    private int AggregateHashes(Symbol operatorSym, IEnumerable<int> hashes) {
      var hashesArray = hashes.ToArray();

      int start;
      if (ReferenceEquals(operatorSym, Addition) && hashesArray.Count() <= 1) {
        start = 0;
      } else {
        start = operatorSym.StringRepresentation.GetHashCode();
      }

      return hashesArray.Aggregate(start, (result, ti) => ((result << 5) + result) ^ ti.GetHashCode());
    }
    #endregion

    #region Parse to SymbolicExpressionTree
    public SymbolicExpressionTree ParseSymbolicExpressionTree(SymbolString sentence) {
      Debug.Assert(sentence.Any(), "Trying to evaluate empty sentence!");
      Debug.Assert(sentence.All(s => s is TerminalSymbol), "Trying to evaluate symbol sequence with nonterminalsymbols!");

      symbolicExpressionGrammar.ConfigureAsDefaultRegressionGrammar();

      var rootNode = rootSy.CreateTreeNode();
      var startNode = startSy.CreateTreeNode();
      rootNode.AddSubtree(startNode);

      Stack<TerminalSymbol> parseStack = new Stack<TerminalSymbol>(sentence.OfType<TerminalSymbol>());
      startNode.AddSubtree(ParseSymbolicExpressionTree(parseStack));

      return new SymbolicExpressionTree(rootNode);
    }

    public ISymbolicExpressionTreeNode ParseSymbolicExpressionTree(Stack<TerminalSymbol> parseStack) {
      TerminalSymbol currentSymbol = parseStack.Pop();

      ISymbolicExpressionTreeNode parsedSubTree = null;

      if (ReferenceEquals(currentSymbol, Addition)) {
        parsedSubTree = addSy.CreateTreeNode();
        parsedSubTree.AddSubtree(ParseSymbolicExpressionTree(parseStack)); // left part
        parsedSubTree.AddSubtree(ParseSymbolicExpressionTree(parseStack)); // right part

      } else if (ReferenceEquals(currentSymbol, Multiplication)) {
        parsedSubTree = mulSy.CreateTreeNode();
        parsedSubTree.AddSubtree(ParseSymbolicExpressionTree(parseStack)); // left part
        parsedSubTree.AddSubtree(ParseSymbolicExpressionTree(parseStack)); // right part

      } else if (ReferenceEquals(currentSymbol, Log)) {
        parsedSubTree = logSy.CreateTreeNode();
        parsedSubTree.AddSubtree(ParseSymbolicExpressionTree(parseStack));

      } else if (ReferenceEquals(currentSymbol, Exp)) {
        parsedSubTree = expSy.CreateTreeNode();
        parsedSubTree.AddSubtree(ParseSymbolicExpressionTree(parseStack));

      } else if (ReferenceEquals(currentSymbol, Sin)) {
        parsedSubTree = sinSy.CreateTreeNode();
        parsedSubTree.AddSubtree(ParseSymbolicExpressionTree(parseStack));

      } else if (Var.VariableTerminalSymbols.Contains(currentSymbol)) {
        VariableTreeNode varNode = (VariableTreeNode)varSy.CreateTreeNode();
        varNode.Weight = 1.0;
        varNode.VariableName = currentSymbol.StringRepresentation;
        parsedSubTree = varNode;
      }

      Debug.Assert(parsedSubTree != null);
      return parsedSubTree;
    }
    #endregion

    #region Parse to Infix string

    public SymbolString PostfixToInfixParser(SymbolString phrase) {
      Stack<Symbol> parseStack = new Stack<Symbol>(phrase);

      return PostfixToInfixSubtreeParser(parseStack);
    }

    private SymbolString PostfixToInfixSubtreeParser(Stack<Symbol> parseStack) {
      Symbol head = parseStack.Pop();

      SymbolString result = new SymbolString();

      if (ReferenceEquals(head, Addition) || ReferenceEquals(head, Multiplication)) {
        // right part
        SymbolString rightPart = PostfixToInfixSubtreeParser(parseStack);
        SymbolString leftPart = PostfixToInfixSubtreeParser(parseStack);

        result.AddRange(leftPart);
        result.Add(head);
        result.AddRange(rightPart);

      } else if (ReferenceEquals(head, Log) || ReferenceEquals(head, Exp) || ReferenceEquals(head, Sin)) {
        result.Add(head);
        result.Add(OpeningBracket);
        result.AddRange(PostfixToInfixSubtreeParser(parseStack));
        result.Add(ClosingBracket);

      } else {
        result.Add(head);
      }
      return result;
    }

    #endregion
  }
}
